@@ -12,22 +12,30 @@ report by exception be used with a less frequent update pushed?
 
 """
 import asyncio
+import io
 import json
 import logging
 import os
-from multiprocessing import Queue
+import statistics
+import sys
+import traceback
 from datetime import datetime
 from enum import Enum
+from multiprocessing import Queue
+from pathlib import Path
 
 import pyshark
-from pyshark.packet.packet import Packet as SharkPacket
 from pyshark.capture.capture import TSharkCrashException
+from pyshark.packet.packet import Packet as SharkPacket
 
-from fieldedge_utilities.logger import get_wrapping_logger
-from fieldedge_utilities.path import clean_path
+_log = logging.getLogger(__name__)
+
+LOCALNET_172 = str(os.getenv('LOCALNET_172', True)) == 'True'
+LOCALNET_192 = str(os.getenv('LOCALNET_192', True)) == 'True'
+DEBUG_PACKET_NUMBER = int(os.getenv('DEBUG_PACKET_NUMBER', 0))
+DEBUG_VERBOSE = str(os.getenv('DEBUG_VERBOSE', False)).lower() == 'true'
 
 
-# Ethernet packet types
 class EthernetProtocol(Enum):
     """Mappings for Ethernet packet types."""
     ETH_TYPE_EDP = 0x00bb  # Extreme Networks Discovery Protocol
@@ -52,40 +60,45 @@ class EthernetProtocol(Enum):
     ETH_TYPE_LLDP = 0x88CC  # Link Layer Discovery Protocol
     ETH_TYPE_TEB = 0x6558  # Transparent Ethernet Bridging
 
-# Common/registered IP protocol ports
-class ApplicationPort(Enum):
-    """Mappings for application layer ports."""
-    TCP_SMTP = 25
-    TCP_HTTP = 80
-    TCP_HTTPS = 443
-    TCP_DNS = 53
-    TCP_FTP = 20
-    TCP_FTPC = 21
-    TCP_TELNET = 23
-    TCP_IMAP = 143
-    TCP_RDP = 3389
-    TCP_SSH = 22
-    TCP_HTTP2 = 8080
-    TCP_MODBUS = 502
-    TCP_MODBUSS = 802
-    TCP_MQTT = 1883
-    TCP_MQTTS = 8883
-    TCP_DOCKERAPI = 2375
-    TCP_DOCKERAPIS = 2376
-    TCP_SRCP = 4303
-    TCP_COAP = 5683
-    TCP_COAPS = 5684
-    TCP_DNP2 = 19999
-    TCP_DNP = 20000
-    TCP_IEC60870 = 2404
-    UDP_SNMP = 161
-    UDP_DNS = 53
-    UDP_DHCPQ = 67
-    UDP_DHCPR = 68
-    UDP_NTP = 123
+
+class KnownTcpPorts(Enum):
+    """Mappings for common registered/known application layer TCP ports."""
+    SMTP = 25
+    HTTP = 80
+    HTTP_TLS = 443
+    DNS = 53
+    FTP = 20
+    FTPC = 21
+    TELNET = 23
+    IMAP = 143
+    RDP = 3389
+    SSH = 22
+    HTTP2 = 8080
+    MODBUS = 502
+    MODBUS_TLS = 802
+    MQTT = 1883
+    MQTT_TLS = 8883
+    MQTT_SOCKET = 9001
+    DOCKERAPI = 2375
+    DOCKERAPI_TLS = 2376
+    SRCP = 4303
+    COAP = 5683
+    COAP_TLS = 5684
+    DNP2 = 19999
+    DNP = 20000
+    IEC60870 = 2404
 
 
-def _get_src_dst(packet: SharkPacket) -> tuple:
+class KnownUdpPorts(Enum):
+    """Mappings for common registered/known application layer TCP ports."""
+    SNMP = 161
+    DNS = 53
+    DHCP_QUERY = 67
+    DHCP_RESPONSE = 68
+    NTP = 123
+
+
+def _get_src_dst(packet: SharkPacket) -> 'tuple[str,str]':
     """Returns the packet source and destination hosts as a tuple.
     
     Args:
@@ -95,9 +108,12 @@ def _get_src_dst(packet: SharkPacket) -> tuple:
         A tuple with (source, destination) IP addresses
     """
     if hasattr(packet, 'arp'):
-        return (packet.arp.src_proto_ipv4, packet.arp.dst_proto_ipv4)
+        return (str(packet.arp.src_proto_ipv4), str(packet.arp.dst_proto_ipv4))
     elif hasattr(packet, 'ip'):
-        return (packet.ip.src, packet.ip.dst)
+        return (str(packet.ip.src), str(packet.ip.dst))
+    elif hasattr(packet, 'ipv6'):
+        # return (str(packet.ipv6.src), str(packet.ipv6.dst))
+        raise NotImplementedError(f'IPV6 unsupported')
     else:
         raise NotImplementedError(f'Unable to determine src/dst'
                                   f' for {packet.highest_layer}')
@@ -112,8 +128,16 @@ def _get_ports(packet: SharkPacket) -> tuple:
     Returns:
         A tuple with (source, destination) ports (TCP or UDP)
     """
-    transport_layer = packet[packet.transport_layer]
-    return (int(transport_layer.srcport), int(transport_layer.dstport))
+    if packet.transport_layer:
+        srcport = int(packet[packet.transport_layer].srcport)
+        dstport = int(packet[packet.transport_layer].dstport)
+    elif hasattr(packet, 'icmp') and packet['icmp'].udp_port:
+        srcport = int(packet['icmp'].udp_srcport)
+        dstport = int(packet['icmp'].udp_dstport)
+    else:
+        raise ValueError('Unable to determine transport'
+                            f' for {packet.highest_layer} packet')
+    return (srcport, dstport)
 
 
 def _get_application(packet: SharkPacket) -> str:
@@ -127,28 +151,46 @@ def _get_application(packet: SharkPacket) -> str:
     Returns:
         A string with the application layer protocol e.g. `TCP_MQTTS`
     """
-    application = 'UNKNOWN'
-    if hasattr(packet[packet.highest_layer], 'app_data_proto'):
-        application = str(packet[packet.highest_layer].app_data_proto).upper()
+    highest_layer: str = packet.highest_layer
+    transport_layer: str = ''
+    if packet.transport_layer:
+        transport_layer = packet.transport_layer
     else:
-        (srcport, dstport) = _get_ports(packet)
-        known_ports = tuple(item.value for item in ApplicationPort)
-        application = str(packet.highest_layer).upper()
-        if srcport in known_ports:
-            application = ApplicationPort(srcport).name
-        elif dstport in known_ports:
-            application = ApplicationPort(dstport).name
-        elif (application == packet.transport_layer):
-            application = f'UNKNOWN_{srcport}_{dstport}'
-            if dstport < srcport:
-                application = f'UNKNOWN_{dstport}_{srcport}'
-    if f'{packet.transport_layer}' not in application:
-        application = f'{packet.transport_layer}_{application}'
+        try:
+            transport_layer = str(packet.layers[2].layer_name).upper()
+        except Exception as err:
+            _log.error(err)
+    application = ''
+    if hasattr(packet[highest_layer], 'app_data_proto'):
+        application = str(packet[highest_layer].app_data_proto).upper()
+    if transport_layer:
+        if not application:
+            (srcport, dstport) = _get_ports(packet)
+            ports_lookup = None
+            if transport_layer == 'TCP':
+                ports_lookup = KnownTcpPorts
+            elif transport_layer == 'UDP':
+                ports_lookup = KnownUdpPorts
+            if ports_lookup:
+                known_ports = tuple(item.value for item in ports_lookup)
+                if srcport in known_ports:
+                    application = ports_lookup(srcport).name
+                elif dstport in known_ports:
+                    application = ports_lookup(dstport).name
+            else:
+                if transport_layer != highest_layer:
+                    application = highest_layer.upper()
+                else:
+                    application += f'{dstport}'
+        application = f'{transport_layer}_{application}'
     # identified workarounds for observed pyshark/tshark app_data_proto
-    if 'HTTP-OVER-TLS' in application:
-        application = application.replace('HTTP-OVER-TLS', 'HTTPS')
-    if packet.highest_layer == 'TLS' and not application.endswith('S'):
-        application = f'{application}S'
+    if not application:
+        application = f'{str(packet.highest_layer).upper()}_UNKNOWN'
+    application = application.replace('HTTP-OVER-TLS', 'HTTP_TLS')
+    if highest_layer == 'TLS' and not application.endswith('_TLS'):
+        application += '_TLS'
+    if not (application.startswith('TCP_') or application.startswith('UDP_')):
+        _log.warning(f'Transport layer unknown for packet {packet.number}')
     return application
 
 
@@ -191,8 +233,40 @@ def is_private_ip(ip_addr: str) -> bool:
     return False
 
 
+def _is_localnet_172(addr: str) -> bool:
+    if (addr.startswith('172.') and
+        int(addr.split('.')[1] in range(16, 31 + 1))):
+        return True
+    return False
+
+
+def _is_localnet_192(addr: str) -> bool:
+    if addr.startswith('192.168.'):
+        return True
+    return False
+
+
+def _is_same_subnet(src: str, dst: str) -> bool:
+    src_parts = src.split('.')
+    dst_parts = dst.split('.')
+    for part in range(0, 3):
+        if src_parts[part] != dst_parts[part]:
+            return False
+    return True
+
+
+def _is_multicast(addr: str) -> bool:
+    MULTICAST_RANGE = (224, 239 + 1)
+    first_octet = int(addr.split('.')[0])
+    if first_octet == 255:
+        return True
+    elif first_octet in range(MULTICAST_RANGE[0], MULTICAST_RANGE[1]):
+        return True
+    return False
+
+
 def _is_local_traffic(packet: SharkPacket) -> bool:
-    """Returns true if the source and destination addresses are on the LAN.
+    """Returns true if the source is on the LAN and destinations are cast.
     
     Args:
         packet: A pyshark Packet capture
@@ -201,11 +275,73 @@ def _is_local_traffic(packet: SharkPacket) -> bool:
         True if both addresses are in the LAN range 192.168.x.y 
     """
     src, dst = _get_src_dst(packet)
-    if ((src.startswith('192.168.') and dst.startswith('192.168.'))):
+    if LOCALNET_172:
+        if (_is_localnet_172(src) and _is_localnet_172(dst) and
+            _is_same_subnet(src, dst)):
+            # if DEBUG_VERBOSE:
+            #     _log.debug(f'Local LAN packet {src} -> {dst}')
+            return True
+        if ((_is_localnet_172(src) and _is_multicast(dst)) or
+            (_is_multicast(src) and _is_localnet_172(dst))):
+            # if DEBUG_VERBOSE:
+            #     _log.debug(f'Local multicast packet {src} -> {dst}')
+            return True
+    if LOCALNET_192:
+        if (_is_localnet_192(src) and _is_localnet_192(dst) and
+            _is_same_subnet(src, dst)):
+            # if DEBUG_VERBOSE:
+            #     _log.debug(f'Local LAN packet {src} -> {dst}')
+            return True
+        if ((_is_localnet_192(src) and _is_multicast(dst)) or
+            (_is_multicast(src) and _is_localnet_192(dst))):
+            # if DEBUG_VERBOSE:
+            #     _log.debug(f'Local multicast packet {src} -> {dst}')
+            return True
+    if _is_multicast(src) and _is_multicast(dst):
+        # if DEBUG_VERBOSE:
+        #     _log.debug(f'Local multicast packet {src} -> {dst}')
         return True
     return False
 
 
+def _is_tcp_reset(packet: SharkPacket) -> bool:
+    # TODO: look for RST flag
+    raise NotImplementedError
+
+
+def _check_flags(packet: SharkPacket) -> 'dict|None':
+    # TODO: tcp.analysis.flags && !tcp.analysis.window_update && !tcp.analysis.keep_alive && !tcp.analysis.keep_alive_ack
+    # TODO: tcp.analysis.retransmission
+    IGNORE = [
+        'analysis_flags',
+    ]
+    if hasattr(packet, 'tcp') and hasattr(packet.tcp, 'analysis_flags'):
+        bad_packet = {}
+        for attr in dir(packet.tcp):
+            if attr.startswith('analysis_') and attr not in IGNORE:
+                bad_packet[attr] = getattr(packet.tcp, attr)
+        return bad_packet
+
+
+def _clean_path(pathname: str) -> str:
+    """Adjusts relative and shorthand filenames for OS independence.
+    
+    Args:
+        pathname: The full path/to/file
+    
+    Returns:
+        A clean file/path name for the current OS and directory structure.
+    """
+    if pathname.startswith('$HOME/'):
+        pathname = pathname.replace('$HOME', str(Path.home()), 1)
+    elif pathname.startswith('~/'):
+        pathname = pathname.replace('~', str(Path.home()), 1)
+    if os.path.isdir(os.path.dirname(pathname)):
+        return os.path.realpath(pathname)
+    else:
+        raise ValueError(f'Directory {os.path.dirname(pathname)} not found')
+
+
 class SimplePacket:
     """A simplified packet representation.
     
@@ -220,6 +356,7 @@ class SimplePacket:
         dst (str): Destination IP address
         srcport (int): Source port
         dstport (int): Destination port
+        bad_packet (dict): Metadata present if the packet is suspected to be bad
 
     """
     def __init__(self, packet: SharkPacket, parent_hosts: tuple) -> None:
@@ -227,12 +364,21 @@ class SimplePacket:
         self.timestamp = round(float(packet.sniff_timestamp), 3)
         self.size = int(packet.length)
         self.transport = packet.transport_layer
+        if packet.transport_layer:
+            self.transport = packet.transport_layer
+            self.stream_id = str(packet[self.transport].stream)
+        elif hasattr(packet, 'icmp') and packet['icmp'].udp_port:
+            self.transport = 'UDP'
+            self.stream_id = str(packet['icmp'].udp_stream)
+        else:
+            raise ValueError('Unable to determine transport'
+                                f' for {packet.highest_layer} packet')
         self.src, self.dst = _get_src_dst(packet)
-        self.srcport = int(packet[self.transport].srcport)
-        self.dstport = int(packet[self.transport].dstport)
+        self.srcport, self.dstport = _get_ports(packet)
         self.highest_layer = str(packet.highest_layer).upper()
         self.application = _get_application(packet)
         self.a_b = True if self.src == self._parent_hosts[0] else False
+        self.bad_packet = _check_flags(packet)
 
 
 class Conversation:
@@ -246,13 +392,20 @@ class Conversation:
         stream_id: The stream ID from the tshark capture
         transport: The transport used e.g. TCP, UDP
         ports: A list of transport ports used e.g. [1883]
+        start_ts: The unix timestamp of the first packet sent
         packets: A list of all the packets summarized
         packet_count: The size of the packets list
         bytes_total: The total number of bytes in the conversation
+        bad_packet_count: The number of suspected bad packets
+            (includes retransmit)
+        bytes_bad: The byte count of the suspected bad packets
+            (includes retransmitted bytes)
+        retransmit_count: The number of suspected retransmitted packets
+        bytes_retransmit: The total retransmitted bytes
 
     """
-    def __init__(self, packet: SharkPacket = None, log: logging.Logger = None):
-        self._log = log or get_wrapping_logger()
+    def __init__(self, packet: SharkPacket = None, number: int = None):
+        self.number = number
         self.application: str = None
         self.hosts: tuple = None
         self.a_b: int = 0
@@ -260,9 +413,14 @@ class Conversation:
         self.stream_id: str = None
         self.transport: str = None
         self.ports: list = []
-        self.packets: list[SimplePacket] = []
+        self.packets: 'list[SimplePacket]' = []
         self.packet_count: int = 0
+        self.bad_packet_count: int = 0
+        self.retransmit_count: int = 0
         self.bytes_total: int = 0
+        self.bytes_bad: int = 0
+        self.bytes_retransmit: int = 0
+        self.start_ts: float = None
         if packet is not None:
             self.packet_add(packet)
     
@@ -281,12 +439,19 @@ class Conversation:
         if self.hosts is None:
             return False
         (src, dst) = _get_src_dst(packet)
-        try:
+        if _is_local_traffic(packet):
+            return False
+        stream_id = None
+        if packet.transport_layer:
             transport = packet.transport_layer
-            stream_id = packet[transport].stream
-        except AttributeError as err:
-            self._log.exception(f'{err}')
+            try:
+                stream_id = packet[transport].stream
+            except AttributeError as err:
+                _log.exception(f'{err}')
+        elif hasattr(packet, 'icmp') and packet['icmp'].udp_stream:
+            stream_id = packet['icmp'].udp_stream
         if (src in self.hosts and dst in self.hosts and
+            stream_id is not None and
             stream_id == self.stream_id):
             return True
         return False
@@ -310,51 +475,63 @@ class Conversation:
         if self.hosts is None:
             self.hosts = _get_src_dst(packet)
         elif not(self.is_packet_in_flow(packet)):
+            _log.warning(f'Packet {packet.number} not in flow {self.number}')
             return False
-        (src, dst) = _get_src_dst(packet)
-        del dst   #:Unused
-        if src == self.hosts[0]:
+        try:
+            simple_packet = SimplePacket(packet, self.hosts)
+        except Exception as err:
+            _log.error(err)
+            raise err
+        isotime = datetime.utcfromtimestamp(simple_packet.timestamp).isoformat()[0:23]
+        if DEBUG_VERBOSE:
+            _log.debug(f'Adding packet {packet.number}'
+                       f' to conversation {self.number or 0}:'
+                       f'{isotime}|{simple_packet.application}|'
+                       f'({simple_packet.transport}.{simple_packet.stream_id}'
+                       f':{simple_packet.dstport})'
+                       f'|{simple_packet.size} bytes'
+                       f'|{simple_packet.src}-->{simple_packet.dst}')
+        if simple_packet.src == self.hosts[0]:
             self.a_b += 1
         else:
             self.b_a += 1
         if self.transport is None:
-            if packet.transport_layer is None:
-                err = f'Packet missing transport_layer'
-                self._log.error(err)
-                raise ValueError(err)
-            self.transport = packet.transport_layer
-        elif packet.transport_layer != self.transport:
-            err = (f'Expected transport {self.transport}'
-                f' got {packet.transport_layer}')
-            self._log.error(err)
-            raise ValueError(err)
-        srcport = int(packet[self.transport].srcport)
-        if srcport not in self.ports:
-            self.ports.append(srcport)
-        dstport = int(packet[self.transport].dstport)
-        if dstport not in self.ports:
-            self.ports.append(dstport)
-        stream_id = packet[self.transport].stream
+            self.transport = simple_packet.transport
+        if simple_packet.srcport not in self.ports:
+            self.ports.append(simple_packet.srcport)
+        if simple_packet.dstport not in self.ports:
+            self.ports.append(simple_packet.dstport)
         if self.stream_id is None:
-            self.stream_id = stream_id
-        elif stream_id != self.stream_id:
-            err = (f'Expected stream {self.stream_id}'
-                f' got {packet[self.transport].stream}')
-            self._log.error(err)
+            self.stream_id = simple_packet.stream_id
+        elif simple_packet.stream_id != self.stream_id:
+            err = (f'Packet {packet.number} expected stream {self.stream_id}'
+                   f' but got {simple_packet.stream_id}')
+            _log.error(err)
             raise ValueError(err)
         self.packet_count += 1
-        self.bytes_total += int(packet.length)
+        self.bytes_total += simple_packet.size
+        if self.start_ts is None:
+            self.start_ts = simple_packet.timestamp
         try:
-            simple_packet = SimplePacket(packet, self.hosts)
+            if simple_packet.bad_packet:
+                if 'analysis_retransmission' in simple_packet.bad_packet:
+                    self.retransmit_count += 1
+                    self.bytes_retransmit += simple_packet.size
+                self.bad_packet_count += 1
+                self.bytes_bad += simple_packet.size
+                if DEBUG_VERBOSE:
+                    _log.debug(f'Bad packet {packet.number}'
+                               f' ({simple_packet.bad_packet})')
             self.packets.append(simple_packet)
             if self.application is None:
                 self.application = simple_packet.application
             elif self.application != simple_packet.application:
-                self._log.warning(f'Expected application {self.application}'
-                    f' but got {simple_packet.application}')
+                _log.warning(f'Packet {packet.number}'
+                             f' expected application {self.application}'
+                             f' but got {simple_packet.application}')
             return True
         except Exception as err:
-            self._log.exception(err)
+            _log.exception(err)
             raise err
         
     @staticmethod
@@ -408,6 +585,17 @@ class Conversation:
             series.append(datapoint)
         return series
 
+    def data_series_packet_size_good_bad(self) -> 'tuple[list, list]':
+        good_series = []
+        bad_series = []
+        for packet in self.packets:
+            datapoint = (packet.timestamp, packet.size)
+            if not packet.bad_packet:
+                good_series.append(datapoint)
+            else:
+                bad_series.append(datapoint)
+        return (good_series, bad_series)
+
     def group_packets_by_size(self) -> tuple:
         """Creates dictionaries keyed by similar packet size and direction.
         
@@ -449,8 +637,7 @@ class Conversation:
         packets_a_b, packets_b_a = self.group_packets_by_size()
         # TODO: dominant packet list based on quantity * size
         return {
-            'A': self.hosts[0],
-            'B': self.hosts[1],
+            'hosts': self.hosts,
             'AB_intervals': self._get_intervals_by_length(packets_a_b),
             'BA_intervals': self._get_intervals_by_length(packets_b_a)
         }
@@ -466,21 +653,23 @@ class PacketStatistics:
 
     """
     def __init__(self,
-                 log: logging.Logger = None,
                  source_filename: str = None,
                  ) -> None:
         """Creates a PacketStatistics object.
         
         Args:
-            log: An optional logging facility
             source_filename: An optional tie to the source pcap file
 
         """
-        self._log = log or get_wrapping_logger()
         self._source_filename: str = source_filename
         self.conversations: list[Conversation] = []
         self._packet_count: int = 0
+        self._unhandled_packet_types: list = []
+        self._unhandled_packet_count: int = 0
+        self._local_packet_count: int = 0
         self._bytes_total: int = 0
+        self._unhandled_bytes: int = 0
+        self._local_bytes: int = 0
         self._first_packet_ts: float = None
         self._last_packet_ts: float = None
     
@@ -512,9 +701,7 @@ class PacketStatistics:
 
         """
         self._packet_count += 1
-        packet_type = packet.highest_layer
-        packet_length = int(packet.length)
-        self._bytes_total += packet_length
+        self._bytes_total += int(packet.length)
         ts = round(float(packet.sniff_timestamp), 3)
         if self._first_packet_ts is None:
             self._first_packet_ts = ts
@@ -523,48 +710,59 @@ class PacketStatistics:
             self._process_arp(packet)
         elif hasattr(packet, 'tcp') or hasattr(packet, 'udp'):
             self._process_ip(packet)
+        elif hasattr(packet, 'icmp'):
+            self._process_ip(packet)
         else:
-            self._log.warning(f'Unhandled packet type {packet_type}')
-            return
+            self._process_unhandled(packet)
     
     def _process_arp(self, packet: SharkPacket):
         arp_desc = f'{packet.arp.src_proto_ipv4}-->{packet.arp.dst_proto_ipv4}'
         if not _is_local_traffic(packet):
-            self._log.warning(f'Non-local ARP packet {arp_desc}')
+            _log.warning(f'Non-local ARP packet {arp_desc}')
         else:
-            self._log.debug(f'Local ARP {arp_desc} (ignored from statistics)')
+            if DEBUG_VERBOSE:
+                _log.debug(f'Local ARP {arp_desc} (ignored from statistics)')
 
     def _process_ip(self, packet: SharkPacket):
         in_conversation = False
+        if _is_local_traffic(packet):
+            if DEBUG_VERBOSE:
+                _log.debug(f'Ignoring packet {packet.number} local traffic')
+            self._local_packet_count += 1
+            self._local_bytes += int(packet.length)
+            return
         for conversation in self.conversations:
             if conversation.is_packet_in_flow(packet):
                 conversation.packet_add(packet)
                 in_conversation = True
+                break
         if not in_conversation:
-            self._log.debug('Found new conversation')
-            conversation = Conversation(packet, self._log)
+            conversation_number = len(self.conversations) + 1
+            if DEBUG_VERBOSE:
+                _log.debug(f'Found new conversation ({conversation_number})')
+            conversation = Conversation(packet, conversation_number)
             self.conversations.append(conversation)
-        packet_type = packet.highest_layer
-        packet_length = int(packet.length)
-        ts = round(float(packet.sniff_timestamp), 3)
-        transport = packet.transport_layer
-        if transport is not None:
-            # srcport = packet[transport].srcport   # unused
-            dstport = packet[transport].dstport
-            stream_id = packet[transport].stream
-        isotime = datetime.utcfromtimestamp(ts).isoformat()[0:23]
-        self._log.debug(f'{isotime}|{packet_type}|'
-            f'({transport}.{stream_id}:{dstport})'
-            f'|{packet_length} bytes|{packet.ip.src}-->{packet.ip.dst}')
 
-    def data_series_application_size(self) -> dict:
+    def _process_unhandled(self, packet: SharkPacket):
+        packet_type = packet.highest_layer
+        self._unhandled_packet_count += 1
+        self._unhandled_bytes += int(packet.length)
+        if packet_type not in self._unhandled_packet_types:
+            _log.warning(f'Packet {packet.number}'
+                         f' unhandled packet type {packet_type}')
+            self._unhandled_packet_types.append(packet_type)
+
+    def data_series_application_size(self, split_bad: bool = False) -> dict:
         """Returns a set of data series by conversation application.
         
         Example: {'MQTT': [(12345.67, 42)]}
 
+        Args:
+            split_bad: if True will split out bad packets as a series.
+
         Returns:
             A dictionary with keys showing the application and values are
-                tuples with (unix_timestamp, size_bytes)
+                tuples with (unix_timestamp, size_in_bytes)
 
         """
         multi_series = {}
@@ -578,34 +776,92 @@ class PacketStatistics:
             multi_series[app].sort(key=lambda tup: tup[0])
         return multi_series
 
-    def analyze_conversations(self) -> list:
-        """Analyzes the conversations.
+    def analyze_conversations(self) -> dict:
+        """Analyzes all conversations to produce a summary.
         
         Returns:
-            A list of analyses currently consisting of the intervals between
-                similarly sized packets of a given application.
+            A dict with keys as unique host pairs "('A', 'B')" summary dict:
+                {
+                    count: `int`,
+                    applications: `list[str]`,
+                    start_times: `list[float]`,
+                    packet_intervals: {
+                        AB_intervals: {
+                            '<transport>_<protocol>_<bytesize>': `int`|`None`,
+                        },
+                        BA_intervals: {
+                            '<transport>_<protocol>_<bytesize>': `int`|`None`,
+                        }
+                    },
+                    repeat_mean: `int`,
+                    repeat_stdev: `int`
+                }
 
         """
+        results = {}
+        for conversation in self.conversations:
+            hosts_str = str(conversation.hosts)
+            intervals = conversation.intervals()
+            intervals.pop('hosts', None)
+            bad_packet_count = conversation.bad_packet_count
+            if hosts_str not in results:
+                results[hosts_str] = {
+                    'count': 1,
+                    'applications': [conversation.application],
+                    'start_times': [conversation.start_ts],
+                    'packet_intervals': intervals,
+                    'bytes': conversation.bytes_total,
+                    'bad_packet_count': bad_packet_count,
+                    'bytes_bad': conversation.bytes_bad,
+                    'retransmit_count': conversation.retransmit_count,
+                    'retransmit_bytes': conversation.bytes_retransmit,
+                }
+            else:
+                results[hosts_str]['count'] += 1
+                app = conversation.application
+                if app not in results[hosts_str]['applications']:
+                    results[hosts_str]['applications'].append(app)
+                results[hosts_str]['start_times'].append(conversation.start_ts)
+                prior = results[hosts_str]['packet_intervals']
+                results[hosts_str]['packet_intervals'] = {**prior, **intervals}
+                results[hosts_str]['bad_packet_count'] += bad_packet_count
+        for key in results:
+            times = results[key]['start_times']
+            results[key]['repeat_mean'] = None
+            results[key]['repeat_stdev'] = None
+            if len(times) == 1:
+                continue
+            intervals = []
+            for i, ts in enumerate(times):
+                if i == 0:
+                    continue
+                intervals.append(ts - times[i - 1])
+            if len(intervals) > 1:
+                results[key]['repeat_mean'] = int(statistics.mean(intervals))
+                results[key]['repeat_stdev'] = int(statistics.stdev(intervals))
+        return results
+    
+    def unique_host_pairs(self) -> 'list[tuple]':
+        """Lists unique host pairs as tuples."""
         results = []
         for conversation in self.conversations:
-            analysis = conversation.intervals()
-            # self._log.info(analysis)
-            results.append(analysis)
+            if conversation.hosts not in results:
+                results.append(conversation.hosts)
         return results
 
 
 def _get_event_loop() -> tuple:
-    newloop = False
+    loop_is_new = False
     try:
-        loop = asyncio.get_running_loop()
+        loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
     except RuntimeError as err:
         if 'no running event loop' not in f'{err}':
             raise err
         loop = asyncio.new_event_loop()
-        newloop = True
+        loop_is_new = True
     asyncio.set_event_loop(loop)
     asyncio.get_child_watcher().attach_loop(loop)
-    return loop, newloop
+    return loop, loop_is_new
 
 
 def process_pcap(filename: str,
@@ -649,33 +905,53 @@ def process_pcap(filename: str,
         A PacketStatistics object with data and analytics functions.
 
     """
-    log = get_wrapping_logger()
-    packet_stats = PacketStatistics(log=log, source_filename=filename)
-    file = clean_path(filename)
-    loop = None
-    newloop = False
+    packet_stats = PacketStatistics(source_filename=filename)
+    file = _clean_path(filename)
+    loop: asyncio.AbstractEventLoop = None
+    loop_is_new = False
     if queue is not None:
-        loop, newloop = _get_event_loop()
+        loop, loop_is_new = _get_event_loop()
     capture = pyshark.FileCapture(input_file=file,
-        display_filter=display_filter, eventloop=loop)
+                                  display_filter=display_filter,
+                                  eventloop=loop)
     capture.set_debug(debug)
     packet_number = 0
+    handled_exceptions = []
     for packet in capture:
+        assert isinstance(packet, SharkPacket)
         packet_number += 1
+        # DEV: Uncomment below for specific step-through troubleshooting
+        if DEBUG_PACKET_NUMBER and packet_number == DEBUG_PACKET_NUMBER:
+            _log.info(f'Investigate packet: {int(packet.number)}')
         try:
             packet_stats.packet_add(packet)
         except NotImplementedError as err:
-            log.error(f'pyshark: {err}')
+            if str(err) not in handled_exceptions:
+                sio = io.StringIO()
+                ei = sys.exc_info()
+                tb = ei[2]
+                traceback.print_exception(ei[0], ei[1], tb, None, sio)
+                s = sio.getvalue()
+                sio.close()
+                stack = s.split('\n')
+                #: each stack call is 2 lines, the last 2 lines are the error
+                #:   so the last call meta is 4-deep and the call itself 3-deep
+                last_call = stack[-4:-2]
+                err_prefix = 'pyshark'
+                if 'fieldedge_pcap/pcap.py' in last_call[0]:
+                    err_prefix = 'fieldedge_pcap'
+                _log.warning(f'{err_prefix} (packet {packet.number}): {err}')
+                handled_exceptions.append(str(err))
         except TSharkCrashException as err:
-            log.error(f'{err}')
+            _log.error(f'tshark (packet {packet_number}): {err}')
             break
-        except Exception as err:
+        except:
             #TODO: better error capture e.g. appears to have been cut short use editcap
             # https://tshark.dev/share/pcap_preparation/
-            log.exception(f'Packet {packet_number} processing ERROR:\n{err}')
+            _log.exception(f'Packet {packet_number} processing ERROR')
             break
     capture.close()
-    if newloop:
+    if loop_is_new:
         loop.close()
     if queue is not None:
         queue.put(packet_stats)
@@ -683,7 +959,7 @@ def process_pcap(filename: str,
         return packet_stats
 
 
-def pcap_filename(duration: int) -> str:
+def pcap_filename(duration: int, interface: str = '') -> str:
     """Generates a pcap filename using datetime of the capture start.
     
     The datetime is UTC, and the duration is in seconds.
@@ -693,8 +969,8 @@ def pcap_filename(duration: int) -> str:
 
     """
     dt = datetime.utcnow().isoformat().replace('-', '').replace(':', '')[0:15]
-    filename = f'capture_{dt}_{duration}.pcap'
-    return filename
+    filename = f'capture_{dt}_{duration}' + f'_{interface}' if interface else ''
+    return f'{filename}.pcap'
 
 
 def create_pcap(interface: str = 'eth1',
@@ -703,7 +979,6 @@ def create_pcap(interface: str = 'eth1',
                 target_directory: str = '$HOME',
                 queue: Queue = None,
                 debug: bool = False,
-                log: logging.Logger = None,
                 ) -> str:
     """Creates a packet capture file of a specified interface.
 
@@ -718,8 +993,8 @@ def create_pcap(interface: str = 'eth1',
     kwargs = {
         'interface': my_interface,
         'duration': my_duration,
-        'filename': pcap_filename(),
-        'target_directory': my_folder,
+        'filename': pcap_filename(duration),
+        'target_directory': parent_folder,
         'queue': queue,
     }
     capture_process = multiprocessing.Process(target=create_pcap,
@@ -748,24 +1023,24 @@ def create_pcap(interface: str = 'eth1',
         The full file/path name if no event is passed in.
 
     """
-    log = log or get_wrapping_logger()
     if filename is None:
         filename = pcap_filename(duration)
-    target_directory = clean_path(target_directory)
+    target_directory = _clean_path(target_directory)
     subdir = f'{target_directory}/{filename[0:len("capture_YYYYmmdd")]}'
     filepath = f'{subdir}/{filename}'
     if not os.path.isdir(subdir):
         os.makedirs(subdir)
-    loop = None
-    newloop = False
+    loop: asyncio.AbstractEventLoop = None
+    loop_is_new = False
     if queue is not None:
-        loop, newloop = _get_event_loop()
-    capture = pyshark.LiveCapture(interface=interface, output_file=filepath,
-        eventloop=loop)
+        loop, loop_is_new = _get_event_loop()
+    capture = pyshark.LiveCapture(interface=interface,
+                                  output_file=filepath,
+                                  eventloop=loop)
     capture.set_debug(debug)
     capture.sniff(timeout=duration)
     capture.close()
-    if newloop:
+    if loop_is_new:
         loop.close()
     if queue is not None:
         queue.put(filepath)
